import math
import os
import random
from collections import defaultdict
from typing import List, Tuple, Union

import cv2
import numpy as np
from tqdm import tqdm

import supervisely as sly
from dataset_tools.image.stats.basestats import BaseVisual
from dataset_tools.image.renders.convert import from_mp4_to_webm, compress_mp4, compress_png
from supervisely.imaging import font as sly_font

UNLABELED_COLOR = [0, 0, 0]
font_name = "FiraSans-Regular.ttf"


class ClassesPreview(BaseVisual):
    """Get previews for classes in the dataset."""

    def __init__(
        self,
        project_meta: sly.ProjectMeta,
        project_name: str,
        api: sly.Api = None,
        row_height: int = None,
        force: bool = False,
    ):
        self.force = force
        self._meta = project_meta
        self._project_name = project_name
        classes_cnt = len(self._meta.obj_classes)
        classes_text = "classes" if classes_cnt > 1 else "class"
        self._title = f"{self._project_name} · {classes_cnt} {classes_text}"

        self._gap = 20
        self._img_width = 1920
        self._img_height = None
        self._row_height = row_height if row_height is not None else 480
        self._row_width = None

        self._api = api if api is not None else sly.Api.from_env()

        self._classname2images = defaultdict(list)
        self._np_images = {}
        self._np_anns = {}
        self._np_texts = {}

        self._logo_path = "logo.png"

    def update(self, image: sly.ImageInfo, ann: sly.Annotation) -> None:
        for label in ann.labels:
            image_area = image.width * image.height
            if image_area * 0.8 < label.area < image_area * 0.1:
                continue
            class_name = label.obj_class.name
            self._classname2images[class_name].append((image, ann))

    def animate(self, path: str = None):
        dirname = os.path.dirname(path)
        os.makedirs(dirname, exist_ok=True)
        self._collect_images()
        canvas, masks, texts = self._prepare_layouts()

        duration, fps = 1.1, 15
        num_frames = int(duration * fps)
        frames = []
        num_frames_list = [0] * 10
        num_frames_list.extend(list(range(0, num_frames)))
        num_frames_list.extend([num_frames] * 10)
        num_frames_list.extend(list(range(num_frames, 0, -1)))

        for i in num_frames_list:
            alpha = i / num_frames
            if i == num_frames:
                alpha = round(1 - random.uniform(0, 0.03), 3)
            elif i == 0:
                alpha = round(0 + random.uniform(0, 0.03), 3)
            frame = self._overlay_images(canvas, masks, alpha)
            frame = self._overlay_images(frame, texts, 1)
            frame = self._add_logo(frame)
            frame = self._draw_title(frame, self._title)
            frames.append(frame)

        tmp_video_path = f"{os.path.splitext(path)[0]}-o.mp4"
        video_path = f"{os.path.splitext(path)[0]}.mp4"
        self._save_video(tmp_video_path, frames)
        from_mp4_to_webm(tmp_video_path, path)
        compress_mp4(tmp_video_path, video_path)
        sly.fs.silent_remove(tmp_video_path)

        sly.logger.info(f"Animation saved to: {path}, {video_path}")

    def _prepare_layouts(self):
        canvas = self._create_grid(list(self._np_images.values()))
        masks = self._create_grid(list(self._np_anns.values()))
        texts = self._create_grid(list(self._np_texts.values()))

        alpha = np.zeros((self._img_height, self._row_width), dtype=canvas.dtype)
        canvas = cv2.merge([canvas, alpha])
        masks = cv2.merge([masks, alpha])
        texts = cv2.merge([texts, alpha])
        masks[:, :, 3] = np.where(np.all(masks == 0, axis=-1), 0, 255).astype(np.uint8)
        texts[:, :, 3] = np.where(np.all(texts == 0, axis=-1), 0, 255).astype(np.uint8)

        return canvas, masks, texts

    def _collect_images(self) -> None:
        with tqdm(
            desc="Download and prepare images and annotations",
            total=len(self._classname2images),
        ) as pbar:
            for cls_name, items in self._classname2images.items():
                random.shuffle(items)
                items: List[Tuple[sly.ImageInfo, sly.Annotation]]
                items = sorted(
                    items,
                    key=lambda item: max(
                        [
                            label.area / (item[0].width * item[0].height)
                            for label in item[1].labels
                            if label.obj_class.name == cls_name
                        ]
                    ),
                )
                image, ann = items[len(items) // 2]

                image_area = image.width * image.height
                img = self._api.image.download_np(image.id, keep_alpha=True)
                pad = "10%"
                crops = sly.aug.instance_crop(
                    img=img,
                    ann=ann,
                    class_title=cls_name,
                    save_other_classes_in_crop=False,
                    padding_config={"top": pad, "bottom": pad, "left": pad, "right": pad},
                )
                crops = sorted(crops, key=lambda crop: crop[1].labels[0].area / image_area)
                cropped_img, cropped_ann = crops[-1]

                cropped_img = self._resize_image(cropped_img, self._row_height)
                cropped_ann = cropped_ann.resize(cropped_img.shape[:2])
                ann_mask = np.zeros((*cropped_img.shape[:2], 3), dtype=np.uint8)
                text_mask = np.zeros((*cropped_img.shape[:2], 3), dtype=np.uint8)
                cropped_ann.draw_pretty(ann_mask, thickness=5, opacity=1)

                label = cropped_ann.labels[0]
                bbox = label.geometry.to_bbox()
                f_scale = self._get_optimal_font_scale(cls_name, (bbox.height, bbox.width))
                font = cv2.FONT_HERSHEY_SIMPLEX
                t_width, _ = cv2.getTextSize(cls_name, font, f_scale, thickness=3)[0][:2]

                if type(label.geometry) in [sly.Bitmap, sly.Polygon]:
                    col, row = bbox.center.col, bbox.center.row
                    org = (col - int(t_width / 2), row)
                    white = (255, 255, 255, 255)
                    cv2.putText(text_mask, cls_name, org, font, f_scale, white, 3, cv2.LINE_AA)

                else:
                    color = label.obj_class.color
                    cv2.rectangle(
                        ann_mask, (bbox.left, bbox.top), (bbox.right, bbox.bottom), color, 2
                    )
                    org = (bbox.top, bbox.left)
                    white = (255, 255, 255, 255)
                    cv2.putText(text_mask, cls_name, org, font, f_scale, white, 3, cv2.LINE_AA)
                self._np_images[cls_name] = cropped_img
                self._np_anns[cls_name] = ann_mask
                self._np_texts[cls_name] = text_mask
                pbar.update(1)

    def _resize_image(self, image: np.ndarray, height: int) -> np.ndarray:
        img_h, img_w = image.shape[:2]
        img_aspect_ratio = height / img_h
        img_h = height
        img_w = int(img_aspect_ratio * img_w)
        image = sly.image.resize(image, (img_h, img_w))

        return image

    def _create_grid(self, images: List[np.ndarray], channels: int = 3) -> np.ndarray:
        rows = self._create_rows(images)
        rows = self._merge_img_in_rows(rows, channels)
        canvas = self._merge_rows_in_canvas(rows, channels)

        return canvas

    def _create_rows(self, images: List[np.ndarray]) -> List[List[np.ndarray]]:
        num_images = len(images)
        rows, _ = self._get_grid_size(num_images)
        self._img_height = rows * (self._row_height + self._gap) + self._gap
        image_widths = [image.shape[1] for image in images]
        
        if rows == 1:
            one_big_row_width = sum(image_widths) + (num_images - 1) * self._gap
        else:
            one_big_row_width = sum(image_widths[:-rows]) + (num_images - 1) * self._gap
        self._row_width = one_big_row_width // rows
        if num_images == 1:
            one_big_row_width = images[0].shape[1]
            self._row_width = one_big_row_width
            return [images]

        rows = []
        row_images = []
        current_width = 0

        for idx, (image, width) in enumerate(zip(images, image_widths)):
            if current_width + width > self._row_width:
                rows.append(row_images)

                row_images = []
                current_width = 0
            row_images.append(image)
            if idx == num_images - 1:
                rows.append(row_images)
            current_width += width + self._gap

        return rows

    def _get_grid_size(self, num: int = 1, aspect_ratio: Union[float, int] = 1.9) -> tuple:
        cols = max(int(math.sqrt(num) * aspect_ratio), 1)
        rows = max((num - 1) // cols + 1, 1)
        return (rows, cols)

    def _merge_img_in_rows(self, rows: List[np.ndarray], channels: int = 3) -> List[np.ndarray]:
        max_row_width = 0
        img_widths = []
        for row in rows:
            sum_widths = sum([img.shape[1] for img in row])
            img_widths.append(sum_widths)
            min_gap = self._gap * (len(row) + 1)
            max_row_width = max(max_row_width, sum_widths + min_gap)
        self._row_width = max_row_width

        combined_rows = []
        for row, width in zip(rows, img_widths):
            if len(row) == 0:
                continue
            if len(row) == 1:
                return row
            gap = (self._row_width - width) // max((len(row) - 1), 1)
            separator = np.ones((self._row_height, gap, channels), dtype=np.uint8) * 255
            combined_images = []

            for image in row:
                combined_images.append(image)
                combined_images.append(separator)
            combined_images.pop()
            combined_image = np.hstack(combined_images)
            combined_rows.append(combined_image)

        return combined_rows

    def _merge_rows_in_canvas(self, rows, channels: int = 3) -> np.ndarray:
        canvas = np.ones([self._img_height, self._row_width, channels], dtype=np.uint8) * 255
        for i, image in enumerate(rows):
            if image.shape[1] != canvas.shape[1] - self._gap * 2:
                image = image[:, : canvas.shape[1] - self._gap * 2]

            row_start = i * (self._row_height + self._gap) + self._gap
            row_end = row_start + self._row_height
            column_start = self._gap
            column_end = image.shape[1] + self._gap

            if row_end > canvas.shape[0]:
                continue
            canvas[row_start:row_end, column_start:column_end] = image

        return canvas

    def _get_optimal_font_scale(self, text: str, image_size: tuple) -> float:
        font_scale = 10
        thickness = 3
        text_height_percent = 10
        font = cv2.FONT_HERSHEY_SIMPLEX
        text_size = cv2.getTextSize(text, font, font_scale, thickness)[0]
        text_width = text_size[0]
        text_height = text_size[1]

        while text_width > image_size[1] * 0.6:
            font_scale -= 0.1
            text_size = cv2.getTextSize(text, font, font_scale, thickness)[0]
            text_width = text_size[0]
            text_height = text_size[1]

        desired_text_height = (image_size[0] * text_height_percent) // 100
        font_scale *= desired_text_height / text_height
        return font_scale

    def _overlay_images(
        self, bg: np.ndarray, overlay: np.ndarray, opacity: Union[float, int]
    ) -> np.ndarray:
        alpha = overlay[..., 3] * opacity / 255.0
        one_minus_alpha = 1 - alpha

        result_image = np.copy(bg)
        for c in range(3):
            result_image[..., c] = bg[..., c] * one_minus_alpha + overlay[..., c] * alpha
        return result_image

    def _add_logo(self, image):
        height = max(self._row_height // 5, 80)
        image2 = cv2.imread(self._logo_path, cv2.IMREAD_UNCHANGED)
        image2 = cv2.cvtColor(image2, cv2.COLOR_BGRA2RGBA)
        image2 = self._resize_image(image2, height)

        height_r = height
        while image2.shape[1] > image.shape[1] * 0.25:
            height_r = int(0.95 * height_r)
            image2 = self._resize_image(image2, height_r)

        h1, w1 = image.shape[:2]
        h2, w2 = image2.shape[:2]
        x = w1 - w2 - self._gap * 2
        y = h1 - h2 - self._gap * 2
        alpha = image2[:, :, 3] / 255.0

        reg = image[y : y + h2, x : x + w2]
        image[y : y + h2, x : x + w2, :3] = (1 - alpha[:, :, np.newaxis]) * reg[:, :, :3] + alpha[
            :, :, np.newaxis
        ] * image2[:, :, :3]

        return image

    def _draw_title(self, image, text):
        image_h, image_w = image.shape[:2]
        font_scale = self._get_optimal_font_scale(text, image.shape[:2])
        font_size = sly_font.get_readable_font_size((image_h, image_w))
        font = sly_font.get_font(font_name, int(font_size * font_scale))
        l, t, r, b = font.getbbox(text)
        title_w, title_h = r - l + self._gap, b
        while title_h > self._row_height * 0.3 or title_w > image_w * 0.8:
            font = font.font_variant(size=int(font.size * 0.96))
            l, t, r, b = font.getbbox(text)
            title_w, title_h = r - l + self._gap, b
        x, y = (image_w - title_w) // 2, self._gap * 2
        left, top, right, bottom = (x, y, x + title_w, y + title_h)
        tmp_canvas = np.zeros((title_h, title_w, 3), dtype=np.uint8)
        tmp_canvas = self._gradient(tmp_canvas, 0, 0, title_w, title_h)
        color = (255, 255, 255, 255)
        sly.image.draw_text(tmp_canvas, text, (-t // 2, self._gap // 2), "tl", font, False, color)
        image[top - 3 : bottom + 3, left - 3 : right + 3, :3] = 255
        image[top:bottom, left:right, :3] = tmp_canvas
        return image

    def _gradient(self, img, left, top, right, bottom):
        c1 = np.array((225, 181, 62))  # rgb
        c2 = np.array((219, 84, 150))  # rgb
        im = np.zeros((bottom - top, right - left, 3), dtype=np.uint8)
        row = np.linspace(0, 1, right - left)
        kernel_1d = np.tile(row, (bottom - top, 1))
        kernel = cv2.merge((kernel_1d, kernel_1d, kernel_1d))
        im = kernel * c1 + (1 - kernel) * c2
        im = im.astype(np.uint8)
        img[top:bottom, left:right, :3] = im
        return img

    def _save_video(self, videopath: str, frames):
        fourcc = cv2.VideoWriter_fourcc(*"VP90")
        height, width = frames[0].shape[:2]
        video_writer = cv2.VideoWriter(videopath, fourcc, 15, (width, height))

        with tqdm(desc="Saving video...", total=len(frames)) as vid_pbar:
            for frame in frames:
                if frame.shape[:2] != (height, width):
                    raise Exception("Not all frame sizes are not equal to each other.")
                frame = cv2.cvtColor(frame, cv2.COLOR_RGBA2BGR)
                video_writer.write(frame)
                vid_pbar.update(1)
        video_writer.release()



import zipfile, os, random, csv
import supervisely as sly
import sly_globals as g
from supervisely.io.fs import get_file_name, get_file_name_with_ext
import numpy as np
import gdown
from cv2 import connectedComponents
from typing import Literal


def download_and_extract():
    gdown.download(g.coffee_url, g.archive_path, quiet=False)
    if zipfile.is_zipfile(g.archive_path):
        with zipfile.ZipFile(g.archive_path, "r") as archive:
            archive.extractall(g.work_dir_path)
    else:
        sly.logger.warn("Archive cannot be unpacked {}".format(g.arch_name))


def read_csv(file_path):
    with open(file_path, newline="") as File:
        reader = csv.reader(File)
        for row in reader:
            g.tags_data[row[0]] = row[1:]


def get_tags(file_id):
    curr_tags = g.tags_data[file_id]
    tags = []
    for idx, curr_tag_val in enumerate(curr_tags):
        if int(curr_tag_val) == 0:
            continue

        tag = sly.Tag(g.tag_metas[idx], value=int(curr_tag_val))
        tags.append(tag)

    return sly.TagCollection(tags)


def create_annotation(ann_path):
    ann_np_leaf = sly.image.read(ann_path)[:, :, g.leaf_idx]
    mask_leaf = ann_np_leaf == g.obj_class_color_idxs[g.leaf_idx]
    bitmap = sly.Bitmap(mask_leaf)
    label = sly.Label(bitmap, g.obj_classes[g.leaf_idx])
    labels = [label]

    ann_np_symptom = sly.image.read(ann_path)[:, :, g.symptom_idx]
    mask_symptom = ann_np_symptom == g.obj_class_color_idxs[g.symptom_idx]
    if len(np.unique(mask_symptom)) != 1:
        ret, curr_mask = connectedComponents(mask_symptom.astype("uint8"), connectivity=8)
        for i in range(1, ret):
            obj_mask = curr_mask == i
            bitmap = sly.Bitmap(obj_mask)
            label = sly.Label(bitmap, g.obj_classes[g.symptom_idx])
            labels.append(label)

    file_id = get_file_name_with_ext(ann_path).split(g.ann_ext)[0]
    tags = get_tags(file_id)

    return sly.Annotation(
        img_size=(ann_np_leaf.shape[0], ann_np_leaf.shape[1]), labels=labels, img_tags=tags
    )


def to_supervisely(api: sly.Api):
    download_and_extract()

    coffee_data_path = os.path.join(g.work_dir_path, sly.fs.get_file_name(g.arch_name))
    tags_file = os.path.join(coffee_data_path, g.symptom_tag_file)
    read_csv(tags_file)

    new_project = api.project.create(g.WORKSPACE_ID, g.project_name, change_name_if_conflict=True)
    api.project.update_meta(new_project.id, g.meta.to_json())

    for ds in g.datasets:
        new_dataset = api.dataset.create(new_project.id, ds, change_name_if_conflict=True)

        curr_img_path = os.path.join(coffee_data_path, g.images_folder, ds.lower())
        curr_ann_path = os.path.join(coffee_data_path, g.anns_folder, ds.lower())

        curr_img_cnt = g.sample_img_count[ds]
        sample_img_path = random.sample(os.listdir(curr_img_path), curr_img_cnt)

        progress = sly.Progress("Create dataset {}".format(ds), curr_img_cnt, sly.logger)
        for img_batch in sly.batched(sample_img_path, batch_size=g.batch_size):
            img_pathes = [os.path.join(curr_img_path, name) for name in img_batch]
            ann_pathes = [
                os.path.join(curr_ann_path, get_file_name(name) + g.ann_ext) for name in img_batch
            ]

            anns = [create_annotation(ann_path) for ann_path in ann_pathes]

            img_infos = api.image.upload_paths(new_dataset.id, img_batch, img_pathes)
            img_ids = [im_info.id for im_info in img_infos]
            api.annotation.upload_anns(img_ids, anns)
            progress.iters_done_report(len(img_batch))

    return new_project.id


def from_supervisely(
    input_path: str, output_path: str = None, to_format: Literal["dir", "tar", "both"] = "both"
) -> str:
    pass



import os, random
import zipfile, json, tarfile
import supervisely as sly
import glob
from supervisely.io.fs import download, file_exists, get_file_name, get_file_name_with_ext
from supervisely.imaging.color import generate_rgb
from pathlib import Path
import numpy as np
from PIL import Image
from tqdm import tqdm
from dotenv import load_dotenv

from typing import Literal
import shutil


from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.polygon import Polygon
from supervisely.io.json import dump_json_file
from supervisely.io.fs import mkdir, get_file_name, get_file_ext, silent_remove
from supervisely.imaging.image import write

from dataset_tools.convert import unpack_if_archive

# my_app = sly.AppService()
# TEAM_ID = int(os.environ["context.teamId"])
# WORKSPACE_ID = int(os.environ["context.workspaceId"])
# INPUT_DIR = os.environ.get("modal.state.slyFolder")
# INPUT_FILE = os.environ.get("modal.state.slyFile")
PROJECT_NAME = "cityscapes"
IMAGE_EXT = ".png"
logger = sly.logger
samplePercent = int(80) / 100
train_tag = "train"
trainval_tag = "trainval"
val_tag = "val"

city_classes_to_colors = {
    "unlabeled": (0, 0, 0),
    "ego vehicle": (98, 15, 138),
    "rectification border": (15, 120, 55),
    "out of roi": (125, 138, 15),
    "static": (63, 15, 138),
    "dynamic": (111, 74, 0),
    "ground": (81, 0, 81),
    "road": (128, 64, 128),
    "sidewalk": (244, 35, 232),
    "parking": (250, 170, 160),
    "rail track": (230, 150, 140),
    "building": (70, 70, 70),
    "wall": (102, 102, 156),
    "fence": (190, 153, 153),
    "guard rail": (180, 165, 180),
    "bridge": (150, 100, 100),
    "tunnel": (150, 120, 90),
    "pole": (153, 153, 153),
    "polegroup": (153, 153, 153),
    "traffic light": (250, 170, 30),
    "traffic sign": (220, 220, 0),
    "vegetation": (107, 142, 35),
    "terrain": (152, 251, 152),
    "sky": (70, 130, 180),
    "person": (220, 20, 60),
    "rider": (255, 0, 0),
    "car": (0, 0, 142),
    "truck": (0, 0, 70),
    "bus": (0, 60, 100),
    "caravan": (0, 0, 90),
    "trailer": (0, 0, 110),
    "train": (0, 80, 100),
    "motorcycle": (0, 0, 230),
    "bicycle": (119, 11, 32),
    "license plate": (0, 0, 142),
}

city_colors = list(city_classes_to_colors.values())


def create_sly_dataset_dir(dst_path, dataset_name):
    dataset_dir = os.path.join(dst_path, dataset_name)
    sly.fs.mkdir(dataset_dir)
    img_dir = os.path.join(dataset_dir, "img")
    sly.fs.mkdir(img_dir)
    ann_dir = os.path.join(dataset_dir, "ann")
    sly.fs.mkdir(ann_dir)
    return dataset_dir


def json_path_to_image_path(json_path):
    img_path = json_path.replace("/gtFine/", "/leftImg8bit/")
    img_path = img_path.replace("_gtFine_polygons.json", "_leftImg8bit" + IMAGE_EXT)
    return img_path


def convert_points(simple_points):
    return [sly.PointLocation(int(p[1]), int(p[0])) for p in simple_points]


def get_split_idxs(num_imgs, percentage):
    train_sample_idxs = int(np.floor(num_imgs * percentage))
    random_idxs = random.sample(population=range(num_imgs), k=train_sample_idxs)
    return random_idxs


def to_supervisely(input_path: str, output_path: str = None):
    input_dir = unpack_if_archive(input_path)

    tag_metas = sly.TagMetaCollection()
    obj_classes = sly.ObjClassCollection()
    dataset_names = []

    if not output_path:
        output_path = os.path.join(os.path.dirname(input_path), "CITYSCAPES_TO_SLY")

    out_project = sly.Project(output_path, sly.OpenMode.CREATE)
    tags_template = os.path.join(input_dir, "gtFine", "*")
    tags_paths = glob.glob(tags_template)
    tags = [os.path.basename(tag_path) for tag_path in tags_paths]
    if train_tag in tags and val_tag not in tags:
        split_train = True
    elif trainval_tag in tags and val_tag not in tags:
        split_train = True
    else:
        split_train = False

    search_fine = os.path.join(input_dir, "gtFine", "*", "*", "*_gt*_polygons.json")
    files_fine = glob.glob(search_fine)
    files_fine.sort()
    search_imgs = os.path.join(input_dir, "leftImg8bit", "*", "*", "*_leftImg8bit" + IMAGE_EXT)
    files_imgs = glob.glob(search_imgs)
    files_imgs.sort()
    if len(files_fine) == 0 or len(files_imgs) == 0:
        raise Exception("Input cityscapes format not correct")

    samples_count = len(files_fine)
    images_pathes_for_compare = []
    images_pathes = {}
    images_names = {}
    anns_data = {}

    if samples_count > 2:
        random_train_indexes = get_split_idxs(samples_count, samplePercent)
    with tqdm(desc="Converting images to sly format", total=samples_count) as pbar:
        for idx, orig_ann_path in enumerate(files_fine):
            parent_dir, json_filename = os.path.split(os.path.abspath(orig_ann_path))
            dataset_name = os.path.basename(parent_dir)
            if dataset_name not in dataset_names:
                dataset_names.append(dataset_name)

                ds = out_project.create_dataset(dataset_name)

                images_pathes[dataset_name] = []
                images_names[dataset_name] = []
                anns_data[dataset_name] = []
            orig_img_path = json_path_to_image_path(orig_ann_path)
            images_pathes_for_compare.append(orig_img_path)
            if not file_exists(orig_img_path):
                logger.warn(
                    "Image for annotation {} not found is dataset {}".format(
                        orig_ann_path.split("/")[-1], dataset_name
                    )
                )
                continue
            images_pathes[dataset_name].append(orig_img_path)
            images_names[dataset_name].append(sly.io.fs.get_file_name_with_ext(orig_img_path))
            tag_path = os.path.split(parent_dir)[0]
            train_val_tag = os.path.basename(tag_path)
            if split_train is True and samples_count > 2:
                if (train_val_tag == train_tag) or (train_val_tag == trainval_tag):
                    if idx in random_train_indexes:
                        train_val_tag = train_tag
                    else:
                        train_val_tag = val_tag

            tag_meta = sly.TagMeta("split", sly.TagValueType.ANY_STRING)
            if not tag_metas.has_key(tag_meta.name):
                tag_metas = tag_metas.add(tag_meta)

            tag = sly.Tag(meta=tag_meta, value=train_val_tag)
            json_data = json.load(open(orig_ann_path))
            ann = sly.Annotation.from_img_path(orig_img_path)

            for obj in json_data["objects"]:
                class_name = obj["label"]
                if class_name == "out of roi":
                    polygon = obj["polygon"][:5]
                    interiors = [obj["polygon"][5:]]
                else:
                    polygon = obj["polygon"]
                    if len(polygon) < 3:
                        logger.warn(
                            "Polygon must contain at least 3 points in ann {}, obj_class {}".format(
                                orig_ann_path, class_name
                            )
                        )
                        continue
                    interiors = []
                interiors = [convert_points(interior) for interior in interiors]
                polygon = sly.Polygon(convert_points(polygon), interiors)
                if city_classes_to_colors.get(class_name, None):
                    obj_class = sly.ObjClass(
                        name=class_name,
                        geometry_type=sly.Polygon,
                        color=city_classes_to_colors[class_name],
                    )
                else:
                    new_color = generate_rgb(city_colors)
                    city_colors.append(new_color)
                    obj_class = sly.ObjClass(
                        name=class_name, geometry_type=sly.Polygon, color=new_color
                    )
                ann = ann.add_label(sly.Label(polygon, obj_class))
                if not obj_classes.has_key(class_name):
                    obj_classes = obj_classes.add(obj_class)
            ann = ann.add_tag(tag)
            anns_data[dataset_name].append(ann)
            ds.add_item_file(sly.fs.get_file_name_with_ext(orig_img_path), orig_img_path, ann=ann)

            pbar.update(1)

    out_meta = sly.ProjectMeta(
        obj_classes=obj_classes, tag_metas=tag_metas, project_type=sly.ProjectType.IMAGES.value
    )
    out_project.set_meta(out_meta)

    stat_dct = {
        "samples": samples_count,
        "src_ann_cnt": len(files_fine),
        "src_img_cnt": len(files_imgs),
    }
    logger.info("Found img/ann pairs.", extra=stat_dct)
    images_without_anns = set(files_imgs) - set(images_pathes_for_compare)
    if len(images_without_anns) > 0:
        logger.warn("Found source images without corresponding annotations:")
        for im_path in images_without_anns:
            logger.warn("Annotation not found {}".format(im_path))

    logger.info(
        "Found classes.",
        extra={
            "cnt": len(obj_classes),
            "classes": sorted([obj_class.name for obj_class in obj_classes]),
        },
    )
    logger.info(
        "Created tags.",
        extra={
            "cnt": len(out_meta.tag_metas),
            "tags": sorted([tag_meta.name for tag_meta in out_meta.tag_metas]),
        },
    )

    return output_path


# RESULT_DIR_NAME = "cityscapes_format"
images_dir_name = "leftImg8bit"
annotations_dir_name = "gtFine"
default_dir_train = "train"
default_dir_val = "val"
default_dir_test = "test"
cityscapes_images_suffix = "_leftImg8bit"
cityscapes_polygons_suffix = "_gtFine_polygons.json"
cityscapes_color_suffix = "_gtFine_color.png"
cityscapes_labels_suffix = "_gtFine_labelIds.png"
possible_geometries = [Bitmap, Polygon]
possible_tags = ["train", "val", "test"]
splitter_coef = 3 / 5
if splitter_coef > 1 or splitter_coef < 0:
    raise ValueError(
        "train_to_val_test_coef should be between 0 and 1, your data is {}".format(splitter_coef)
    )


def from_supervisely(
    input_path: str, output_path: str = None, to_format: Literal["dir", "tar", "both"] = "both"
) -> str:
    input_dir = unpack_if_archive(input_path)

    project_fs = sly.Project(input_dir, sly.OpenMode.READ)
    meta = project_fs.meta
    datasets = project_fs.datasets

    # api.file.list

    # api = sly.Api()

    storage_dir = os.path.dirname(input_dir)

    if not output_path:
        RESULT_DIR_NAME = "SLY_TO_CITYSCAPES"
        # output_path = os.path.join(os.path.dirname(input_path), "SLY_TO_CITYSCAPES")

    def get_image_and_ann():
        mkdir(image_dir_path)
        mkdir(ann_dir)
        image_path = os.path.join(image_dir_path, image_name)
        # api.image.download_path(image_id, image_path)
        image_ext_to_png(image_path)

        mask_color, mask_label, poly_json = from_ann_to_cityscapes_mask(
            ann, name2id, sly.logger, train_val_flag
        )
        # dump_json_file(poly_json,
        #                os.path.join(ann_dir, get_file_name(base_image_name) + cityscapes_polygons_suffix))
        # write(
        #     os.path.join(ann_dir,
        #                  get_file_name(base_image_name) + cityscapes_color_suffix), mask_color)
        # write(
        #     os.path.join(ann_dir,
        #                  get_file_name(base_image_name) + cityscapes_labels_suffix), mask_label)

        dump_json_file(
            poly_json,
            os.path.join(
                ann_dir,
                get_file_name(base_image_name).replace("_leftImg8bit", "")
                + cityscapes_polygons_suffix,
            ),
        )
        write(
            os.path.join(
                ann_dir,
                get_file_name(base_image_name).replace("_leftImg8bit", "")
                + cityscapes_color_suffix,
            ),
            mask_color,
        )
        write(
            os.path.join(
                ann_dir,
                get_file_name(base_image_name).replace("_leftImg8bit", "")
                + cityscapes_labels_suffix,
            ),
            mask_label,
        )

    ARCHIVE_NAME = "{}_cityscapes.tar".format(project_fs.name)

    has_bitmap_poly_shapes = False
    for obj_class in meta.obj_classes:
        if obj_class.geometry_type not in possible_geometries:
            sly.logger.warn(
                f"Cityscapes format supports only bitmap and polygon classes, {obj_class.geometry_type} will be skipped"
            )
        else:
            has_bitmap_poly_shapes = True

    if has_bitmap_poly_shapes is False:
        raise Exception("Input project does not contain bitmap or polygon classes")

    RESULT_ARCHIVE = os.path.join(storage_dir, ARCHIVE_NAME)
    RESULT_DIR = os.path.join(storage_dir, RESULT_DIR_NAME)
    result_images_train = os.path.join(RESULT_DIR, images_dir_name, default_dir_train)
    result_images_val = os.path.join(RESULT_DIR, images_dir_name, default_dir_val)
    result_images_test = os.path.join(RESULT_DIR, images_dir_name, default_dir_test)
    result_anns_train = os.path.join(RESULT_DIR, annotations_dir_name, default_dir_train)
    result_anns_val = os.path.join(RESULT_DIR, annotations_dir_name, default_dir_val)
    result_anns_test = os.path.join(RESULT_DIR, annotations_dir_name, default_dir_test)
    sly.fs.mkdir(RESULT_DIR)
    sly.logger.info("Cityscapes Dataset folder has been created")

    class_to_id = []
    name2id = {}
    for idx, obj_class in enumerate(meta.obj_classes):
        if obj_class.geometry_type not in possible_geometries:
            continue
        curr_class = {}
        curr_class["name"] = obj_class.name
        curr_class["id"] = idx + 1
        curr_class["color"] = obj_class.color
        class_to_id.append(curr_class)
        name2id[obj_class.name] = (idx + 1, idx + 1, idx + 1)

    dump_json_file(class_to_id, os.path.join(RESULT_DIR, "class_to_id.json"))
    sly.logger.info("Writing classes with colors to class_to_id.json file")

    with tqdm(desc="Convert dataset to original format", total=project_fs.total_items) as pbar:
        for dataset in datasets:
            images_dir_path_train = os.path.join(result_images_train, dataset.name)
            images_dir_path_val = os.path.join(result_images_val, dataset.name)
            images_dir_path_test = os.path.join(result_images_test, dataset.name)
            anns_dir_path_train = os.path.join(result_anns_train, dataset.name)
            anns_dir_path_val = os.path.join(result_anns_val, dataset.name)
            anns_dir_path_test = os.path.join(result_anns_test, dataset.name)

            images = os.listdir(dataset.img_dir)

            if len(images) < 3:
                sly.logger.warn(
                    "Number of images in {} dataset is less then 3, val and train directories for this dataset will not be created".format(
                        dataset.name
                    )
                )

            # image_ids = [image_info.id for image_info in images]
            base_image_names = [sly.fs.get_file_name_with_ext(img) for img in images]
            # image_names = [
            #     get_file_name(image_info.name) + cityscapes_images_suffix + get_file_ext(image_info.name) for
            #     image_info in images
            # ]

            image_names = [
                get_file_name(img.replace("_leftImg8bit", ""))
                + cityscapes_images_suffix
                + get_file_ext(img)
                for img in images
            ]

            anns = [dataset.get_ann(name, meta) for name in image_names]

            splitter = get_tags_splitter(anns)
            curr_splitter = {"train": 0, "val": 0, "test": 0}

            for ann, image_name, base_image_name in zip(anns, image_names, base_image_names):
                train_val_flag = True
                try:
                    split_name = ann.img_tags.get("split").value
                    if split_name == "train":
                        image_dir_path = images_dir_path_train
                        ann_dir = anns_dir_path_train
                    elif split_name == "val":
                        image_dir_path = images_dir_path_val
                        ann_dir = anns_dir_path_val
                    else:
                        image_dir_path = images_dir_path_test
                        ann_dir = anns_dir_path_test
                        train_val_flag = False
                except:
                    ann_tags = [tag.name for tag in ann.img_tags]
                    separator_tags = list(set(ann_tags) & set(possible_tags))
                    if len(separator_tags) > 1:
                        sly.logger.warn(
                            """There are more then one separator tag for {} image. {}
                        tag will be used for split""".format(
                                image_name, separator_tags[0]
                            )
                        )

                    if len(separator_tags) >= 1:
                        if separator_tags[0] == "train":
                            image_dir_path = images_dir_path_train
                            ann_dir = anns_dir_path_train
                        elif separator_tags[0] == "val":
                            image_dir_path = images_dir_path_val
                            ann_dir = anns_dir_path_val
                        else:
                            image_dir_path = images_dir_path_test
                            ann_dir = anns_dir_path_test
                            train_val_flag = False

                    if len(separator_tags) == 0:
                        if curr_splitter["test"] == splitter["test"]:
                            curr_splitter = {"train": 0, "val": 0, "test": 0}
                        if curr_splitter["train"] < splitter["train"]:
                            curr_splitter["train"] += 1
                            image_dir_path = images_dir_path_train
                            ann_dir = anns_dir_path_train
                        elif curr_splitter["val"] < splitter["val"]:
                            curr_splitter["val"] += 1
                            image_dir_path = images_dir_path_val
                            ann_dir = anns_dir_path_val
                        elif curr_splitter["test"] < splitter["test"]:
                            curr_splitter["test"] += 1
                            image_dir_path = images_dir_path_test
                            ann_dir = anns_dir_path_test
                            train_val_flag = False

                get_image_and_ann()
                pbar.update(1)

    if to_format in ("tar", "both"):
        sly.fs.archive_directory(RESULT_DIR, RESULT_ARCHIVE)
        sly.logger.info("Result directory is archived")
        if to_format == "tar":
            shutil.rmtree(RESULT_DIR)

    if to_format == "tar":
        return RESULT_ARCHIVE
    elif to_format == "dir":
        return RESULT_DIR
    else:
        return (RESULT_ARCHIVE, RESULT_DIR)


def get_tags_splitter(anns):
    anns_without_possible_tags = 0
    for ann in anns:
        ann_tags = [tag.name for tag in ann.img_tags]
        separator_tags = list(set(ann_tags) & set(possible_tags))
        if len(separator_tags) == 0:
            anns_without_possible_tags += 1
    train_tags_cnt = round(anns_without_possible_tags * splitter_coef)
    val_tags_cnt = round((anns_without_possible_tags - train_tags_cnt) / 2)
    test_tags_cnt = anns_without_possible_tags - train_tags_cnt - val_tags_cnt
    return {"train": train_tags_cnt, "val": val_tags_cnt, "test": test_tags_cnt}


def image_ext_to_png(im_path):
    if get_file_ext(im_path) != ".png":
        im = Image.open(im_path).convert("RGB")
        im.save(im_path[: -1 * len(get_file_ext(im_path))] + ".png")
        silent_remove(im_path)


def from_ann_to_cityscapes_mask(ann, name2id, app_logger, train_val_flag):
    mask_color = np.zeros((ann.img_size[0], ann.img_size[1], 3), dtype=np.uint8)
    mask_label = np.zeros((ann.img_size[0], ann.img_size[1], 3), dtype=np.uint8)
    poly_json = {"imgHeight": ann.img_size[0], "imgWidth": ann.img_size[1], "objects": []}

    for label in ann.labels:
        if type(label.geometry) not in possible_geometries:
            continue
        if train_val_flag:
            label.geometry.draw(mask_color, label.obj_class.color)
        label.geometry.draw(mask_label, name2id[label.obj_class.name])
        if type(label.geometry) == Bitmap:
            curr_cnt = label.geometry.to_contours()
            if len(curr_cnt) == 0:
                continue
            elif len(curr_cnt) == 1:
                poly_for_contours = curr_cnt[0]
            else:
                for poly in curr_cnt:
                    cur_contours = poly.exterior_np.tolist()
                    if len(poly.interior) > 0 and label.obj_class.name != "out of roi":
                        app_logger.info(
                            "Labeled objects must never have holes in cityscapes format, existing holes will be sketched"
                        )
                    cityscapes_contours = list(map(lambda cnt: cnt[::-1], cur_contours))
                    poly_json["objects"].append(
                        {"label": label.obj_class.name, "polygon": cityscapes_contours}
                    )
                continue
        else:
            poly_for_contours = label.geometry

        if len(poly_for_contours.interior) > 0 and label.obj_class.name != "out of roi":
            app_logger.info(
                "Labeled objects must never have holes in cityscapes format, existing holes will be sketched"
            )

        contours = poly_for_contours.exterior_np.tolist()

        if label.obj_class.name == "out of roi":
            for curr_interior in poly_for_contours.interior_np:
                contours.append(poly_for_contours.exterior_np.tolist()[0])
                contours.extend(curr_interior.tolist())
                contours.append(curr_interior.tolist()[0])

        cityscapes_contours = list(map(lambda cnt: cnt[::-1], contours))
        poly_json["objects"].append({"label": label.obj_class.name, "polygon": cityscapes_contours})

    return mask_color, mask_label, poly_json



import os
from datetime import datetime

import numpy as np
import supervisely as sly


class Previews:
    def __init__(
        self,
        project_id,
        project_meta,
        api,
        team_id,
        force: bool = False,
    ):
        self.project_meta = project_meta
        self.project_id = project_id
        self.force = force

        self.MAX_WIDTH = 500
        self.BATCH_SIZE = 50

        self.api = api
        self.team_id = team_id

        self.render_dir = f"/dataset/{project_id}/renders"

        self.images_batch = []
        self.errors = []

    def update(self, image, ann):
        self.images_batch.append((image, ann))

        if len(self.images_batch) >= self.BATCH_SIZE:
            self._save_batch()
            self.images_batch.clear()

    def _save_batch(self):
        existing_renders = self.api.file.list2(self.team_id, self.render_dir, recursive=False)
        existing_ids = set(int(sly.fs.get_file_name(f.path)) for f in existing_renders)

        local_paths, remote_paths = [], []

        for image, ann in self.images_batch:
            if image.id in existing_ids:
                continue

            if image.width > self.MAX_WIDTH:
                out_size = (int((image.height / image.width) * self.MAX_WIDTH), self.MAX_WIDTH)
                ann = ann.resize(out_size)

            render = np.zeros((ann.img_size[0], ann.img_size[1], 3), dtype=np.uint8)
            ann.draw(render, thickness=ann._get_thickness())
            alpha = (1 - np.all(render == [0, 0, 0], axis=-1).astype("uint8")) * 255
            rgba = np.dstack((render, alpha))

            local_path = os.path.join(os.getcwd(), "tmp/renders", f"{image.id}.png")
            remote_path = os.path.join(self.render_dir, f"{image.id}.png")

            local_paths.append(local_path)
            remote_paths.append(remote_path)
            sly.image.write(local_path, rgba, remove_alpha_channel=False)

        try:
            self.api.file.upload_bulk(self.team_id, local_paths, remote_paths)
        except Exception as e:
            self.errors.append(
                {
                    "timestamp": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                    "image_ids": [image.id for image, ann in self.images_batch],
                    "error": str(e),
                }
            )

        for path in local_paths:
            sly.fs.silent_remove(path)

    def close(self):
        if len(self.images_batch) > 0:
            self._save_batch()

        if len(self.errors) > 0:
            print("Errors during renders upload:")
            for error in self.errors:
                print(error)



import os, sys
from pathlib import Path
import supervisely as sly

root_source_dir = str(Path(sys.argv[0]).parents[1])
sly.logger.info(f"Root source directory: {root_source_dir}")
sys.path.append(root_source_dir)


TEAM_ID = sly.env.team_id()
WORKSPACE_ID = sly.env.workspace_id()

datasets = ["Train", "Val", "Test"]

sample_img_count = {"Train": 400, "Val": 50, "Test": 50}

project_name = "Coffee Leaf Biotic Stress"
work_dir = "coffee_data"
coffee_url = "https://docs.google.com/uc?id=15YHebAGrx1Vhv8-naave-R5o3Uo70jsm"

arch_name = "coffee-datasets.zip"
images_folder = "segmentation/images"
anns_folder = "segmentation/annotations"
symptom_tag_file = "leaf/dataset.csv"
ann_ext = "_mask.png"

batch_size = 30

obj_classes_names = ["symptom", "leaf"]
symptom_idx = 0
leaf_idx = 1
obj_class_color_idxs = [255, 176]
symptom_color = [255, 0, 0]
leaf_color = [0, 176, 0]

obj_class_leaf = sly.ObjClass(obj_classes_names[1], sly.Bitmap, color=leaf_color)
obj_class_symptom = sly.ObjClass(obj_classes_names[0], sly.Bitmap, color=symptom_color)

obj_classes = [obj_class_symptom, obj_class_leaf]
obj_class_collection = sly.ObjClassCollection(obj_classes)

tags_data = {}
tag_names = ["predominant_stress", "miner", "rust", "phoma", "cercospora", "severity"]
tag_metas = []
for tag_name in tag_names:
    tag_meta = sly.TagMeta(tag_name, sly.TagValueType.ANY_NUMBER)
    tag_metas.append(tag_meta)

tag_meta_collection = sly.TagMetaCollection(tag_metas)

meta = sly.ProjectMeta(obj_classes=obj_class_collection, tag_metas=tag_meta_collection)

storage_dir = sly.app.get_data_dir()
work_dir_path = os.path.join(storage_dir, work_dir)
sly.fs.mkdir(work_dir_path)
archive_path = os.path.join(work_dir_path, arch_name)
